from copy import copy
from typing import TYPE_CHECKING

from metrics_layer.core.exceptions import (
    AccessDeniedOrDoesNotExistException,
    QueryError,
)

from .base import MetricsLayerBase, SQLReplacement
from .field import Field
from .set import Set

if TYPE_CHECKING:
    from metrics_layer.core.model.project import Project


class ZenlyticJoinType:
    left_outer = "left_outer"
    inner = "inner"
    full_outer = "full_outer"
    cross = "cross"
    options = [left_outer, inner, full_outer, cross]


class ZenlyticJoinRelationship:
    many_to_one = "many_to_one"
    one_to_one = "one_to_one"
    one_to_many = "one_to_many"
    many_to_many = "many_to_many"
    options = [many_to_one, one_to_one, one_to_many, many_to_many]


class Join(MetricsLayerBase, SQLReplacement):
    def __init__(self, definition: dict, project) -> None:
        self.project: Project = project
        if "type" not in definition:
            definition["type"] = ZenlyticJoinType.left_outer
        if "relationship" not in definition:
            definition["relationship"] = ZenlyticJoinRelationship.many_to_one

        self.validate(definition)
        super().__init__(definition)

    def replaced_sql_on(self, query_type: str):
        return self.get_replaced_sql_on(self.sql_on, query_type)

    @property
    def name(self):
        return f"join between {self.base_view_name} and {self.join_view_name}"

    def validate(self, definition: dict):
        required_keys = ["base_view_name", "join_view_name", "relationship", "type"]

        for k in required_keys:
            if k not in definition:
                base_view_name, join_view_name = None, None
                if "base_view_name" in definition:
                    base_view_name = definition["base_view_name"]
                if "join_view_name" in definition:
                    join_view_name = definition["join_view_name"]
                raise QueryError(
                    f"Join missing required key {k} in join with view {base_view_name} and {join_view_name}"
                )

        has_sql_on = "sql_on" in definition
        has_fk = "foreign_key" in definition
        has_sql = "sql" in definition
        # Cross join won't need either sql_on or foreign key set
        is_cross_join = definition["type"] == "cross"
        all_join_arguments = [has_sql_on, has_fk, has_sql, is_cross_join]
        no_join_keys = all(not c for c in all_join_arguments)
        multiple_join_keys = sum(all_join_arguments) > 1

        if no_join_keys:
            raise QueryError(
                f"No join arguments found in join {definition['base_view_name']}, "
                "please pass sql_on or foreign_key"
            )

        if multiple_join_keys:
            raise QueryError(
                f"Multiple join arguments found in join {definition['base_view_name']}"
                ", please pass only one of: sql_on, foreign_key"
            )

    def _error(self, element, error, extra: dict = {}):
        line, column = self.line_col(element)
        return {**extra, "message": error, "line": line, "column": column}

    def collect_errors(self):
        errors = []
        if self.foreign_key:
            for view_name in [self.base_view_name, self.join_view_name]:
                try:
                    self.project.get_field(
                        self.foreign_key,
                        view_name=view_name,
                        show_excluded=True,
                    )
                except Exception:
                    errors.append(
                        self._error(
                            self.foreign_key,
                            (
                                f"Could not find field {self.foreign_key} in {self.name} referencing view"
                                f" {view_name}"
                            ),
                        )
                    )
            return errors

        if self.sql_on:
            fields_to_replace = self.fields_to_replace(self.sql_on)

            for field in fields_to_replace:
                _, view_name, column_name = Field.field_name_parts(field)
                if view_name is None:
                    errors.append(
                        self._error(self.sql_on, f"Could not find view for field {field} in {self.name}")
                    )
                    continue

                try:
                    view = self.project.get_view(view_name)
                except Exception:
                    err_msg = f"Could not find view {view_name} in {self.name}"
                    errors.append(self._error(self.sql_on, err_msg))
                    continue

                try:
                    self.project.get_field(column_name, view_name=view.name)
                except Exception:
                    errors.append(
                        self._error(
                            self.sql_on,
                            f"Could not find field {column_name} in {self.name} referencing view {view_name}",
                        )
                    )

        return errors

    def is_valid(self):
        if self.sql_on:
            fields_to_replace = self.fields_to_replace(self.sql_on)

            # The join isn't valid if we can't find an existing view with that name
            for field in fields_to_replace:
                _, view_name, _ = Field.field_name_parts(field)
                if view_name not in self.explore.join_names():
                    err_msg = f"Could not find view {view_name} for {self.name}"
                    print(err_msg)
                    return False
            return True
        is_valid = self.foreign_key is not None or self.type == "cross"
        return is_valid

    def required_views(self):
        if not self.sql_on:
            return [self.base_view_name, self.join_view_name]

        views = []
        for field in self.fields_to_replace(self.sql_on):
            _, view_name, _ = Field.field_name_parts(field)
            views.append(view_name)
        return list(set(views))

    def required_joins(self):
        if not self.sql_on:
            return [self.explore.name, self.name]

        joins = []
        for field in self.fields_to_replace(self.sql_on):
            _, join_name, _ = Field.field_name_parts(field)
            joins.append(join_name)
        return list(set(joins))

    def get_replaced_sql_on(self, sql: str, query_type: str):
        sql_on = copy(sql)
        fields_to_replace = self.fields_to_replace(sql_on)
        for field in fields_to_replace:
            _, view_name, column_name = Field.field_name_parts(field)
            if view_name is None:
                return

            view = self.project.get_view(view_name)

            if view is None:
                return

            table_name = view.name
            field_obj = self.project.get_field(column_name, view_name=table_name)

            if field_obj and table_name:
                sql_condition = field_obj.sql_query(query_type)
                replace_with = sql_condition
            elif table_name:
                replace_with = f"{table_name}.{column_name}"
            else:
                replace_with = column_name

            replace_text = "${" + field + "}"
            sql_on = sql_on.replace(replace_text, replace_with)

        return sql_on

    def field_names(self):
        # This function is for the explore `fields` parameter, to resolve all the sets into field names
        if self.fields is None:
            return self.fields

        set_definition = {
            "name": "NA",
            "fields": self.fields,
            "view_name": self.from_,
        }
        join_set = Set(set_definition, project=self.project)
        return join_set.field_names()

    def join_fields(self, show_hidden: bool, expand_dimension_groups: bool, show_excluded: bool):
        try:
            view = self.project.get_view(self.from_)
        except AccessDeniedOrDoesNotExistException:
            # If the user does not have access to the view, there are obviously no fields to show them
            return []
        fields = view.fields(show_hidden, expand_dimension_groups)
        join_field_names = self.field_names()
        if join_field_names and not show_excluded:
            return [f for f in fields if f.id() in join_field_names]
        return fields



import os
import re
from typing import List, Union

import pandas as pd

from metrics_layer.core.model.definitions import Definitions


class SeedMetricsLayer:
    default_views_path = "views"
    default_models_path = "models"
    default_dashboards_path = "dashboards"

    def __init__(
        self,
        connection=None,
        database=None,
        schema=None,
        table=None,
        profile=None,
        target=None,
        metrics_layer=None,
        run_query_override=None,
    ):
        self.default_model_name = "base_model"
        self.run_query_override = run_query_override
        self.profile_name = profile
        self.target_name = target
        if metrics_layer is None:
            self.metrics_layer = SeedMetricsLayer._init_profile(self.profile_name, self.target_name)
        else:
            self.metrics_layer = metrics_layer
        self.connection = self._init_connection(self.metrics_layer, connection)
        self.database = database if database else self.connection.database
        self._database_is_not_default = database and database != self.connection.database
        self.schema = schema if schema else self.connection.schema
        self.table = table.replace(".sql", "") if table else None
        if schema and table:
            self.location_description = f"table: {self.database}.{self.schema}.{self.table}"
        elif schema:
            self.location_description = f"schema: {self.database}.{self.schema}"
        else:
            self.location_description = f"database: {self.database}"

        self._snowflake_type_lookup = {
            "DATE": "date",
            "DATETIME": "timestamp",
            "TIMESTAMP_TZ": "timestamp",
            "TIMESTAMP_NTZ": "timestamp",
            "TIMESTAMP_LTZ": "timestamp",
            "TIMESTAMP": "timestamp",
            "BOOLEAN": "yesno",
            "FIXED": "number",
            "FLOAT": "number",
            "REAL": "number",
            "TEXT": "string",
            "VARCHAR": "string",
            "CHAR": "string",
            "CHARACTER": "string",
            "STRING": "string",
            "BINARY": "string",
            "VARBINARY": "string",
            "NUMBER": "number",
        }

        self._redshift_type_lookup = {
            "TIME": "timestamp",
            "TIMETZ": "timestamp",
            "TIMESTAMP": "timestamp",
            "TIMESTAMPTZ": "timestamp",
            "DATE": "date",
            "CHAR": "string",
            "VARCHAR": "string",
            "BOOLEAN": "yesno",
            "DOUBLE PRECISION": "number",
            "DOUBLE": "number",
            "PRECISION": "number",
            "REAL": "number",
            "DECIMAL": "number",
            "BIGINT": "number",
        }
        self._databricks_type_lookup = {
            "TIME": "timestamp",
            "TIMESTAMP": "timestamp",
            "TIMESTAMP_NTZ": "timestamp",
            "DATE": "date",
            "STRING": "string",
            "BOOLEAN": "yesno",
            "DOUBLE": "number",
            "FLOAT": "number",
            "DECIMAL": "number",
            "LONG": "number",
            "INT": "number",
            "SMALLINT": "number",
            "BIGINT": "number",
            "TINYINT": "number",
        }
        self._druid_type_lookup = {
            "CHAR": "string",
            "VARCHAR": "string",
            "DECIMAL": "number",
            "FLOAT": "number",
            "REAL": "number",
            "DOUBLE": "number",
            "BOOLEAN": "yesno",
            "TINYINT": "number",
            "SMALLINT": "number",
            "INTEGER": "number",
            "BIGINT": "number",
            "TIMESTAMP": "timestamp",
            "DATE": "date",
        }
        # Also duck db type lookup
        self._postgres_type_lookup = {
            "date": "date",
            "timestamp without time zone": "timestamp",
            "timestamp with time zone": "timestamp",
            "text": "string",
            "boolean": "yesno",
            "bigint": "number",
            "integer": "number",
            "smallint": "number",
            "real": "number",
            "double precision": "number",
            "numeric": "number",
            "float": "number",
            "money": "number",
        }
        self._sql_server_type_lookup = {
            "float": "number",
            "real": "number",
            "bigint": "number",
            "numeric": "number",
            "smallint": "number",
            "decimal": "number",
            "smallmoney": "number",
            "int": "number",
            "tinyint": "number",
            "money": "number",
            "char": "string",
            "varchar": "string",
            "text": "string",
            "date": "date",
            "datetime2": "datetime",
            "smalldatetime": "datetime",
            "datetime": "datetime",
            "time": "datetime",
            "bit": "yesno",
        }
        self._bigquery_type_lookup = {
            "DATE": "date",
            "DATETIME": "timestamp",
            "TIMESTAMP": "timestamp",
            "BOOL": "yesno",
            "FLOAT64": "number",
            "INT64": "number",
            "NUMERIC": "number",
            "STRING": "string",
        }

    def seed(self, auto_tag_searchable_fields: bool = False):
        from metrics_layer.core.parse import ProjectDumper, ProjectLoader

        if self.connection.type not in Definitions.supported_warehouses:
            raise NotImplementedError(
                f"The only data warehouses supported for seeding are {Definitions.supported_warehouses_text}"
            )
        columns_query = self.columns_query()
        data = self.run_query(columns_query)
        data.columns = [c.upper() for c in data.columns]

        if self.connection.type in {Definitions.snowflake, Definitions.databricks}:
            table_query = self.table_query()
            table_data = self.run_query(table_query)
            table_data.columns = [c.upper() for c in table_data.columns]
        else:
            table_data = pd.DataFrame()
        print(f"Got information on all tables and views in {self.location_description}")

        # We need to filter down the whole result to just the chosen schema and table (if applicable)
        if self.schema:
            data = data[data["TABLE_SCHEMA"].str.lower() == self.schema.lower()].copy()
            if not table_data.empty:
                table_data = table_data[table_data["TABLE_SCHEMA"].str.lower() == self.schema.lower()].copy()

        if self.table:
            data = data[data["TABLE_NAME"].str.lower() == self.table.lower()].copy()
            if not table_data.empty:
                table_data = table_data[table_data["TABLE_NAME"].str.lower() == self.table.lower()].copy()
        folder = self._location()
        loader = ProjectLoader(folder)
        project = loader.load()

        current_models = project.models()

        model_name = self.get_model_name(current_models)
        # Each iteration represents either a single table or a single view
        views, tables = [], data["TABLE_NAME"].unique()
        for i, table_name in enumerate(tables):
            column_df = data[data["TABLE_NAME"].str.lower() == table_name.lower()].copy()
            if not table_data.empty:
                table_comment = table_data[table_data["TABLE_NAME"].str.lower() == table_name.lower()][
                    "COMMENT"
                ].values[0]
            else:
                table_comment = None
            schema_name = column_df["TABLE_SCHEMA"].values[0]
            view = self.make_view(
                column_df,
                model_name,
                table_name,
                schema_name,
                table_comment,
                auto_tag_searchable_fields=auto_tag_searchable_fields,
            )
            if self.table:
                progress = ""
            else:
                progress = f"({i + 1} / {len(tables)})"
            print(f"Got information on {table_name} {progress}")
            views.append(view)

        if len(current_models) > 0:
            models = []
        else:
            models = self.make_models()

        model_folder = loader.zenlytic_project.get("model-paths", [self.default_models_path])[0]
        view_folder = loader.zenlytic_project.get("view-paths", [self.default_views_path])[0]

        # Dump the models to yaml files
        dumper = ProjectDumper(models, model_folder, views, view_folder)
        dumper.dump(folder)

        # Add the zenlytic_project.yml file if it doesn't exist yet
        if len(current_models) == 0 and not os.path.exists(os.path.join(folder, "zenlytic_project.yml")):
            zenlytic_project_path = os.path.join(folder, "zenlytic_project.yml")
            project_data = {
                "name": self.connection.name,
                "profile": self.connection.name,
                "model-paths": ["models"],
                "view-paths": ["views"],
                "dashboard-paths": ["dashboards"],
            }
            dumper.dump_yaml_file(project_data, zenlytic_project_path)

    def get_model_name(self, current_models: list):
        if len(current_models) > 0:
            return current_models[0].name
        return self.default_model_name

    def make_models(self):
        model = {
            "version": 1,
            "type": "model",
            "name": self.default_model_name,
            "connection": self.connection.name,
        }
        return [model]

    def make_view(
        self,
        column_data,
        model_name: str,
        table_name: str,
        schema_name: str,
        table_comment: str = None,
        auto_tag_searchable_fields: bool = True,
    ):
        view_name = self.clean_name(table_name)
        fields = self.make_fields(
            column_data,
            schema_name=schema_name,
            table_name=table_name,
            auto_tag_searchable_fields=auto_tag_searchable_fields,
        )
        if self.connection.type in {
            Definitions.snowflake,
            Definitions.redshift,
            Definitions.postgres,
            Definitions.sql_server,
            Definitions.azure_synapse,
            Definitions.duck_db,
            Definitions.databricks,
        }:
            sql_table_name = f"{schema_name}.{table_name}"
            if self._database_is_not_default:
                sql_table_name = f"{self.database}.{sql_table_name}"
        elif self.connection.type == Definitions.druid:
            sql_table_name = f"{schema_name}.{table_name}"
        elif self.connection.type == Definitions.bigquery:
            sql_table_name = f"`{self.database}.{schema_name}.{table_name}`"
        else:
            raise NotImplementedError(f"Unsupported connection type {self.connection.type}")
        view = {
            "version": 1,
            "type": "view",
            "name": view_name,
            "model_name": model_name,
            "sql_table_name": sql_table_name,
            "default_date": next((f["name"] for f in fields if f["field_type"] == "dimension_group"), None),
            "fields": fields,
        }
        if table_comment:
            view["description"] = table_comment
        if view["default_date"] is None:
            view.pop("default_date")
        return view

    def make_fields(self, column_data, schema_name: str, table_name: str, auto_tag_searchable_fields: bool):
        if auto_tag_searchable_fields:
            if schema_name is None:
                raise ValueError("schema_name is required to auto tag searchable fields")
            if table_name is None:
                raise ValueError("table_name is required to auto tag searchable fields")

        fields = []
        searchable_field_candidates = []
        if self.connection.type in {Definitions.snowflake, Definitions.databricks}:
            data_to_iterate = column_data[["COLUMN_NAME", "DATA_TYPE", "COMMENT"]]
        else:
            data_to_iterate = column_data[["COLUMN_NAME", "DATA_TYPE"]]
        for _, row in data_to_iterate.iterrows():
            name = self.clean_name(row["COLUMN_NAME"])
            if self.connection.type == Definitions.snowflake:
                metrics_layer_type = self._snowflake_type_lookup.get(row["DATA_TYPE"], "string")
            elif self.connection.type == Definitions.redshift:
                metrics_layer_type = self._redshift_type_lookup.get(row["DATA_TYPE"], "string")
            elif self.connection.type in {Definitions.postgres, Definitions.duck_db}:
                metrics_layer_type = self._postgres_type_lookup.get(row["DATA_TYPE"], "string")
            elif self.connection.type == Definitions.bigquery:
                metrics_layer_type = self._bigquery_type_lookup.get(row["DATA_TYPE"], "string")
            elif self.connection.type == Definitions.druid:
                metrics_layer_type = self._druid_type_lookup.get(row["DATA_TYPE"], "string")
            elif self.connection.type in {Definitions.sql_server, Definitions.azure_synapse}:
                metrics_layer_type = self._sql_server_type_lookup.get(row["DATA_TYPE"], "string")
            elif self.connection.type == Definitions.databricks:
                metrics_layer_type = self._databricks_type_lookup.get(row["DATA_TYPE"], "string")
            else:
                raise NotImplementedError(f"Unknown connection type: {self.connection.type}")
            # Add quotes for certain db only because we've seen issues with column names with special chars
            if self.connection.type in {
                Definitions.druid,
                Definitions.snowflake,
                Definitions.duck_db,
                Definitions.postgres,
            }:
                column_name = '"' + row["COLUMN_NAME"] + '"'
            else:
                column_name = row["COLUMN_NAME"]
            sql = "${TABLE}." + column_name

            field = {"name": name, "sql": sql}

            if "COMMENT" in row and row["COMMENT"] is not None:
                field["description"] = row["COMMENT"]

            if metrics_layer_type in {"timestamp", "date", "datetime"}:
                field["field_type"] = "dimension_group"
                field["type"] = "time"
                field["timeframes"] = [
                    "raw",
                    "date",
                    "day_of_year",
                    "week",
                    "week_of_year",
                    "month",
                    "month_of_year",
                    "quarter",
                    "year",
                ]
                field["datatype"] = metrics_layer_type
            elif metrics_layer_type == "string" and auto_tag_searchable_fields:
                field["field_type"] = "dimension"
                field["type"] = "string"
                searchable_field_candidates.append(row["COLUMN_NAME"])
            else:
                field["field_type"] = "dimension"
                field["type"] = metrics_layer_type
            fields.append(field)
        if searchable_field_candidates:
            column_cardinalities_query = self.column_cardinalities_query(
                column_names=searchable_field_candidates, schema_name=schema_name, table_name=table_name
            )
            column_cardinalities = self.run_query(query=column_cardinalities_query)

            # Get the column names that have a cardinality of less than 100
            # Note: running the query doesn't preserve the column name cases
            searchable_column_names = [
                col.lower().rsplit("_cardinality", 1)[0]
                for col in column_cardinalities.columns
                if column_cardinalities.loc[0, col] < 100
            ]

            for field in fields:
                if field["sql"].split(".", 1)[1].lower().replace('"', "") in searchable_column_names:
                    field["searchable"] = True
                elif field["field_type"] == "dimension" and field["type"] == "string":
                    field["searchable"] = False

        return fields

    def column_cardinalities_query(
        self,
        column_names: List[str],
        schema_name: str,
        table_name: str,
        custom_sql_base: Union[str, None] = None,
        quote: bool = True,
        alias_mappings: Union[dict, None] = None,
    ) -> str:
        cardinality_queries = []
        for column_name in column_names:
            if alias_mappings is not None and column_name in alias_mappings:
                column_name_alias = alias_mappings[column_name]
            else:
                column_name_alias = column_name
            if self.connection.type in (Definitions.snowflake, Definitions.duck_db, Definitions.druid):
                quote_column_name = f'"{column_name}"' if quote else column_name
                query = f'APPROX_COUNT_DISTINCT( {quote_column_name} ) as "{column_name_alias}_cardinality"'  # noqa: E501
            elif self.connection.type in {Definitions.redshift, Definitions.postgres}:
                quote_column_name = f'"{column_name}"' if quote else column_name
                query = (
                    f'COUNT(DISTINCT {quote_column_name} ) as "{column_name_alias}_cardinality"'  # noqa: E501
                )
            elif self.connection.type in {Definitions.sql_server, Definitions.azure_synapse}:
                quote_column_name = f'"{column_name}"' if quote else column_name
                query = f'COUNT(DISTINCT cast({quote_column_name} as VARCHAR(MAX))) as "{column_name_alias}_cardinality"'  # noqa: E501
            elif self.connection.type == Definitions.bigquery:
                quote_column_name = f"`{column_name}`" if quote else column_name
                query = (
                    f"APPROX_COUNT_DISTINCT( TO_JSON_STRING({quote_column_name}) ) as"
                    f" `{column_name_alias}_cardinality`"
                )
            elif self.connection.type == Definitions.databricks:
                quote_column_name = f"`{column_name}`" if quote else column_name
                query = f"APPROX_COUNT_DISTINCT( {quote_column_name} ) as `{column_name_alias}_cardinality`"
            else:
                raise NotImplementedError(f"Unknown connection type: {self.connection.type}")
            cardinality_queries.append(query)

        query = f"SELECT {', '.join(cardinality_queries)}"

        if custom_sql_base:
            query += f" FROM {custom_sql_base}"
        elif self.connection.type in {
            Definitions.snowflake,
            Definitions.duck_db,
            Definitions.druid,
            Definitions.redshift,
            Definitions.postgres,
            Definitions.sql_server,
            Definitions.azure_synapse,
            Definitions.databricks,
        }:
            query += f" FROM {self.database}.{schema_name}.{table_name}"
        elif self.connection.type == Definitions.druid:
            query += f"FROM {schema_name}.{table_name}"
        elif self.connection.type == Definitions.bigquery:
            query += f" FROM `{self.database}`.`{schema_name}`.`{table_name}`"

        return query + ";" if self.connection.type != Definitions.druid else query

    def columns_query(self):
        if self.connection.type in {Definitions.snowflake, Definitions.databricks}:
            comment_statement = ", comment as comment"
        else:
            comment_statement = ""
        query = (
            f"SELECT table_catalog, table_schema, table_name, column_name, data_type{comment_statement} FROM "
        )
        if self.database and self.connection.type in {
            Definitions.snowflake,
            Definitions.redshift,
            Definitions.postgres,
            Definitions.sql_server,
            Definitions.azure_synapse,
            Definitions.duck_db,
            Definitions.databricks,
        }:
            if self.connection.type == Definitions.databricks and self.database is None:
                query += f"INFORMATION_SCHEMA.COLUMNS"
            else:
                query += f"{self.database}.INFORMATION_SCHEMA.COLUMNS"
        elif self.connection.type == Definitions.druid:
            query = (
                "SELECT TABLE_CATALOG, TABLE_SCHEMA, TABLE_NAME, COLUMN_NAME, DATA_TYPE "
                "FROM INFORMATION_SCHEMA.COLUMNS"
            )
        elif self.database and self.schema and self.connection.type == Definitions.bigquery:
            query += f"`{self.database}.{self.schema}`.INFORMATION_SCHEMA.COLUMNS"
        elif not self.schema and self.connection.type == Definitions.bigquery:
            raise ValueError(
                "You must specify a database (project) AND a schema (dataset) for seeding in BigQuery"
            )
        else:
            raise ValueError("You must specify at least a database for seeding")

        # Apply where clause if we are seeding a single table
        if self.table and self.schema:
            query += f" WHERE TABLE_SCHEMA = '{self.schema}' AND TABLE_NAME = '{self.table}'"
        elif not self.table and self.schema:
            query += f" WHERE TABLE_SCHEMA = '{self.schema}'"

        # Snowflake had a metadata error when not using the limit statement, so we add this
        if self.connection.type == Definitions.snowflake:
            # 10k columns is a reasonable max for a single table
            return query + " LIMIT 10000;"
        return query + ";" if self.connection.type != Definitions.druid else query

    def table_query(self):
        if self.database and self.connection.type == Definitions.snowflake:
            query = (
                "SELECT table_catalog as table_database, table_schema as table_schema, "
                "table_name as table_name, table_owner as table_owner, table_type as table_type, "
                "bytes as table_size, created as table_created, last_altered as table_last_modified, "
                "row_count as table_row_count, comment as comment "
                f"FROM {self.database}.INFORMATION_SCHEMA.TABLES"
            )
        elif self.connection.type in {Definitions.druid}:
            query = (
                "SELECT TABLE_CATALOG as table_database, TABLE_SCHEMA as table_schema, "
                "TABLE_NAME as table_name, TABLE_TYPE as table_type "
                "FROM INFORMATION_SCHEMA.TABLES "
                "WHERE TABLE_SCHEMA not in ('sys', 'INFORMATION_SCHEMA')"
            )
        elif self.database and self.connection.type in {
            Definitions.redshift,
            Definitions.postgres,
            Definitions.sql_server,
            Definitions.azure_synapse,
            Definitions.duck_db,
        }:
            query = (
                "SELECT table_catalog as table_database, table_schema as table_schema, "
                "table_name as table_name, table_type as table_type "
                f"FROM {self.database}.INFORMATION_SCHEMA.TABLES "
                "WHERE table_schema not in ('pg_catalog', 'information_schema')"
            )
        elif self.connection.type == Definitions.databricks:
            database_str = f"{self.database}." if self.database else ""
            query = (
                "SELECT table_catalog as table_database, table_schema as table_schema, "
                "table_name as table_name, table_type as table_type, "
                f"comment as comment FROM {database_str}INFORMATION_SCHEMA.TABLES"
            )
        elif self.database and self.schema and self.connection.type == Definitions.bigquery:
            query = (
                "SELECT table_catalog as table_database, table_schema as table_schema, "
                "table_name as table_name, table_type as table_type, "
                "creation_time as table_created FROM "
                f"`{self.database}.{self.schema}`.INFORMATION_SCHEMA.TABLES"
            )
        elif not self.schema and self.connection.type == Definitions.bigquery:
            raise ValueError(
                "You must specify a database (project) AND a schema (dataset) for seeding in BigQuery"
            )
        else:
            raise ValueError("You must specify at least a database for seeding")
        return query + ";" if self.connection.type != Definitions.druid else query

    def run_query(self, query: str):
        if self.run_query_override:
            return self.run_query_override(query)
        return self.metrics_layer.run_query(
            query, self.connection, run_pre_queries=False, start_warehouse=True
        )

    @staticmethod
    def _init_connection(metrics_layer, connection_name: str = None):
        if connection_name:
            connection = metrics_layer.get_connection(connection_name)
        else:
            connections = metrics_layer.list_connections()
            if len(connections) == 1:
                connection = connections[0]
            else:
                raise ValueError(
                    f"Could not determine the connection to use, "
                    f"please pass the connection name with the --connection arg"
                )
        return connection

    @staticmethod
    def _init_profile(profile_name: str, target: str = None):
        from metrics_layer.core import MetricsLayerConnection

        connections = MetricsLayerConnection.get_connections_from_profile(profile_name, target)
        metrics_layer = MetricsLayerConnection(location=SeedMetricsLayer._location(), connections=connections)
        metrics_layer.load()
        return metrics_layer

    @staticmethod
    def get_profile():
        from metrics_layer.core.parse.project_reader_base import ProjectReaderBase

        zenlytic_project_path = os.path.join(SeedMetricsLayer._location(), "zenlytic_project.yml")
        dbt_project_path = SeedMetricsLayer._dbt_project_path()

        if os.path.exists(zenlytic_project_path):
            zenlytic_project = ProjectReaderBase.read_yaml_file(zenlytic_project_path)
            return zenlytic_project["profile"]
        elif os.path.exists(dbt_project_path):
            dbt_project = ProjectReaderBase.read_yaml_file(dbt_project_path)
            return dbt_project["profile"]
        raise ValueError(
            """Could not find a profile for the metrics layer in either the zenlytic_project.yml file or in
         the dbt_project.yml file, if neither of those files exist, please create the file"""
        )

    @staticmethod
    def _dbt_project_path():
        return os.path.join(SeedMetricsLayer._location(), "./dbt_project.yml")

    @staticmethod
    def _in_dbt_project():
        return os.path.exists(SeedMetricsLayer._dbt_project_path())

    def _dbt_project_file():
        from metrics_layer.core.parse.project_reader_base import ProjectReaderBase

        if SeedMetricsLayer._in_dbt_project():
            return ProjectReaderBase.read_yaml_file(SeedMetricsLayer._dbt_project_path())
        return None

    @staticmethod
    def _init_directories():
        models_dir = SeedMetricsLayer.default_models_path
        views_dir = SeedMetricsLayer.default_views_path
        dashboards_dir = SeedMetricsLayer.default_dashboards_path

        in_dbt_project = SeedMetricsLayer._in_dbt_project()
        if in_dbt_project:
            to_create = [dashboards_dir]
        else:
            to_create = [models_dir, views_dir, dashboards_dir]

        # Create models and views directory inside of project dir
        for directory in to_create:
            fully_qualified_path = os.path.join(SeedMetricsLayer._location(), directory)
            if not os.path.exists(fully_qualified_path):
                os.mkdir(fully_qualified_path)

    @staticmethod
    def _init_project_file():
        from metrics_layer.core.parse import ProjectDumper

        common = {"dashboard-paths": [SeedMetricsLayer.default_dashboards_path]}
        if SeedMetricsLayer._in_dbt_project():
            dbt_project = SeedMetricsLayer._dbt_project_file()
            default_profile = {"name": dbt_project["name"], "profile": dbt_project["profile"], "mode": "dbt"}
        else:
            default_profile = {
                "name": "zenlytic_project_name",
                "profile": "my_dbt_profile",
                "model-paths": [SeedMetricsLayer.default_models_path],
                "view-paths": [SeedMetricsLayer.default_views_path],
            }

        default_profile = {**default_profile, **common}
        ProjectDumper.dump_yaml_file(default_profile, "zenlytic_project.yml")

    @staticmethod
    def _location():
        return os.getcwd()

    @staticmethod
    def _test_git():
        try:
            import git  # noqa

            valid = True
        except ImportError:
            valid = False
        return valid

    @staticmethod
    def clean_name(txt: str):
        alphanumeric_string = re.sub(r"[^a-zA-Z0-9\s_]", "", txt)
        return alphanumeric_string.lower().strip().replace(" ", "_")



from metricflow_to_zenlytic.metricflow_to_zenlytic import (
    load_mf_project,
    convert_mf_project_to_zenlytic_project,
)
from .project_reader_base import ProjectReaderBase
from metrics_layer.core.exceptions import QueryError


class MetricflowProjectReader(ProjectReaderBase):
    def load(self) -> None:
        if self.dbt_project is None:
            raise QueryError("No dbt project found")

        self.project_name = self.dbt_project["name"]

        metricflow_project = load_mf_project(self.dbt_folder)

        dbt_profile_name = self.dbt_project.get("profile", self.project_name)
        if self.zenlytic_project:
            self.profile_name = self.zenlytic_project.get("profile", dbt_profile_name)
        else:
            self.profile_name = dbt_profile_name

        models, views = convert_mf_project_to_zenlytic_project(
            metricflow_project, self.profile_name, self.profile_name
        )
        return models, views, []



from typing import TYPE_CHECKING, Union

from metrics_layer.core.exceptions import QueryError

from .base import MetricsLayerBase

if TYPE_CHECKING:
    from metrics_layer.core.model.project import Project


class Set(MetricsLayerBase):
    valid_properties = ["name", "fields", "view_name"]

    def __init__(self, definition: dict, project) -> None:
        self.validate(definition)

        self.project: Project = project
        super().__init__(definition)

    def validate(self, definition: dict):
        required_keys = ["name", "fields"]
        for k in required_keys:
            if k not in definition:
                raise QueryError(f"Set missing required key {k}")

    def _error(self, element, error, extra: dict = {}):
        line, column = self.line_col(element)
        error = {**extra, "message": error, "line": line, "column": column}
        if self.view_name:
            error["view_name"] = self.view_name
        return error

    def collect_errors(self):
        errors = []
        if not self.valid_name(self.name):
            errors.append(self._error(self.name, self.name_error("set", self.name)))

        if "view_name" in self._definition and self.valid_name(self.view_name):
            try:
                self.project.get_view(self.view_name)
            except Exception as e:
                errors.append(self._error(self.view_name, f"In the Set {self.name} " + str(e)))

        if not isinstance(self.fields, list):
            errors.append(
                self._error(
                    self._definition["fields"],
                    f"The fields property, {self.fields} must be a list in the Set {self.name}",
                )
            )
        else:
            try:
                field_names = self.field_names()
            except Exception:
                # These errors are handled in the view in which the set is defined
                field_names = []

            for field in field_names:
                try:
                    self.project.get_field(field)
                except Exception as e:
                    errors.append(
                        self._error(self._definition["fields"], f"In the Set {self.name} " + str(e))
                    )

        errors.extend(
            self.invalid_property_error(
                self._definition, self.valid_properties, "set", self.name, error_func=self._error
            )
        )
        return errors

    def field_names(self):
        all_field_names, names_to_exclude = [], []
        for field_name in self.fields:
            # This means we're subtracting a set from the result
            if "*" in field_name and "-" in field_name:
                clean_name = field_name.replace("*", "").replace("-", "")
                fields_to_remove = self._internal_get_fields_from_set(clean_name)
                names_to_exclude.extend(fields_to_remove)

            # This means we're expanding a set into this set
            elif "*" in field_name:
                clean_name = field_name.replace("*", "")
                fields_to_add = self._internal_get_fields_from_set(clean_name)
                all_field_names.extend(fields_to_add)

            # This means we're removing a field from the result
            elif "-" in field_name:
                _, view_name, field_name = self.field_name_parts(field_name.replace("-", ""))
                view_name = self._get_view_name(view_name, field_name)
                names_to_exclude.append(f"{view_name}.{field_name}")

            # This is just a field that we're adding to the result
            else:
                _, view_name, field_name = self.field_name_parts(field_name)
                view_name = self._get_view_name(view_name, field_name)
                all_field_names.append(f"{view_name}.{field_name}")

        # Perform exclusion
        result_field_names = set(f for f in all_field_names if f not in names_to_exclude)

        return sorted(list(result_field_names), key=all_field_names.index)

    def _internal_get_fields_from_set(self, set_name: str):
        if set_name == "ALL_FIELDS":
            all_fields = self.project.fields(
                view_name=self.view_name, show_hidden=True, expand_dimension_groups=True
            )
            return [f"{f.view.name}.{f.alias()}" for f in all_fields]

        _, view_name, set_name = self.field_name_parts(set_name)
        _set = self.project.get_set(set_name, view_name=view_name)
        if _set is None:
            print(f"WARNING: Could not find set with name {set_name}, disregarding those fields")
            return []
        return _set.field_names()

    def _get_view_name(self, view_name: Union[None, str], field_name: str):
        if view_name:
            return view_name
        elif view_name is None and self.view_name:
            return self.view_name
        else:
            raise QueryError(f"Cannot find a valid view name for the field {field_name} in set {self.name}")
